package org.jboss.seam.cache;

import org.jboss.seam.log.LogProvider;
import org.jboss.seam.log.Logging;

import java.io.*;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

/**
 * File bases cache with no transaction guaranties. It can only cache text files and no binaries.
 * It actually only ensures that there is no concurrent access to the 
 * same file. In addition there is an in memory boost cache holding frequent accessed files.
 *
 * You can specify the root folder of the cache, the default cache region and the size of the in memory cache.
 *
 * The cache reports statistics if enabled.
 *
 * @author Sebastian Hennebrueder
 */

public class FileSystemCache {

  private static final LogProvider log = Logging.getLogProvider(FileSystemCache.class);

  /**
   * root directory below all files are stored
   */
  private File root;

  /**
   * Cache of filenames, key and value is the filename
   */
  private ConcurrentMap<String, String> filenameMap = new ConcurrentHashMap<String, String>();
  /**
   * Name of the default cache region
   */
  private String defaultRegion;

  /**
   * Size of a buffer holding cached objects in memory - improve performance of frequent loaded objects
   */
  private int bufferSize = 0;

  /**
   * buffer holding frequent content in memory
   */
  private Map<String, String> boostCache;

  private CacheStatistics cacheStatistics;

  /**
   * Suffix of cached files
   */
  public static final String SUFFIX = ".cache";

  public FileSystemCache(File root, String defaultRegion, int bufferSize) {
    this.root = root;
    this.defaultRegion = defaultRegion;
    this.bufferSize = bufferSize;
    RestrictedLinkedHashMap<String, String> map = new RestrictedLinkedHashMap<String, String>();
    map.setMaxEntries(bufferSize);
    boostCache = Collections.synchronizedMap(map);

  }

  public void setCacheStatistics(CacheStatistics statistics) {
    this.cacheStatistics = statistics;
  }

  public CacheStatistics getCacheStatistics() {
    return cacheStatistics;
  }

  public static class RestrictedLinkedHashMap<K, V> extends LinkedHashMap<K, V> {
    private int maxEntries = 0;

    public void setMaxEntries(int maxEntries) {
      this.maxEntries = maxEntries;
    }


    @Override
    protected boolean removeEldestEntry(Map.Entry<K, V> eldest) {
      return size() > maxEntries;
    }
  }


  /**
   * Returns the existing filename to allow synchronization on this filename when the file is written.
   * Adds the passed filename to the fileNames and returns the new instance if none exist already.
   *
   * @param filename
   * @return
   */
  protected synchronized String getExistingOrMe(String filename) {
    String existingOrMe = filenameMap.putIfAbsent(filename, filename);
    if (existingOrMe == null)
      return filename;
    else
      return existingOrMe;
  }

  public String readFile(String cachedFile) {
    return readFile(null, cachedFile);
  }

  public String readFile(String region, String cachedFilename) {
    String pathToCachedFile = buildCachedFileName(region, cachedFilename);
    String existingOrMe = getExistingOrMe(pathToCachedFile);
    // we synchronize on the same file name
    synchronized (existingOrMe) {
      String content = boostCache.get(existingOrMe);
      if (content != null) {
        // move entry to beginning
        boostCache.remove(existingOrMe);
        boostCache.put(existingOrMe, content);
        writeStatistic(CacheStatistics.EntryType.BOOST_CACHE_HIT);
        return content;
      }
      File cachedFile = new File(existingOrMe);
      if (cachedFile.exists()) {
        StringBuilder buf = new StringBuilder();
        try {
          BufferedReader reader = new BufferedReader(
                  new FileReader(cachedFile));
          int count = 0;
          char[] cbuf = new char[0xfff];
          while ((count = reader.read(cbuf)) != -1) {
            buf.append(cbuf, 0, count);
          }
          reader.close();
          content = buf.toString();
          boostCache.put(existingOrMe, content);
          writeStatistic(CacheStatistics.EntryType.CACHE_HIT);
          return content;
        } catch (FileNotFoundException e) {
          if (log.isDebugEnabled())
            log.debug("Could not read file though it should exist.", e);
        } catch (IOException e) {
          if (log.isDebugEnabled())
            log.debug("Could not read file though it should exist.", e);
        }
      }
      filenameMap.remove(pathToCachedFile);
      writeStatistic(CacheStatistics.EntryType.MISS);
      return null;
    }
  }

  private void writeStatistic(CacheStatistics.EntryType entryType) {
    if (cacheStatistics != null)
      cacheStatistics.addEntry(entryType);
  }

  public void createOrUpdateFile(String key, String content) {
    createOrUpdateFile(null, key, content);
  }

  public void createOrUpdateFile(String region, String key, String content) {
    String pathToCachedFile = buildCachedFileName(region, key);
    String existingOrMe = getExistingOrMe(pathToCachedFile);
    // we synchronize on the same file name
    synchronized (existingOrMe) {
      try {
        BufferedWriter writer = new BufferedWriter(new FileWriter(existingOrMe, false));
        writer.write(content);
        writer.close();

        // move entry to beginning
        if (bufferSize > 0) {
          boostCache.remove(existingOrMe);
          boostCache.put(existingOrMe, content);
        }
      } catch (FileNotFoundException e) {
        // we might be here because the region folder was missing
        String regionFolder = existingOrMe.substring(0, existingOrMe.lastIndexOf("/"));
        File file = new File(regionFolder);
        if (file.mkdir()) {
          // try again
          try {
            BufferedWriter writer = new BufferedWriter(new FileWriter(existingOrMe, false));
            writer.write(content);
            writer.close();
          } catch (IOException e2) {
            log.error(String.format("Error writing file %s to FileSystemCache", existingOrMe), e2);
          }
        } else
          throw new CacheException("Could not create region - " + (region == null ? defaultRegion : region));


      } catch (IOException e) {
        log.error(String.format("Error writing file %s to FileSystemCache", existingOrMe), e);
      }
    }
  }

  public void removeFile(String key) {
    removeFile(null, key);
  }

  public void removeFile(String region, String key) {
    String pathToCachedFile = buildCachedFileName(region, key);
    removeFileByName(pathToCachedFile);
  }

  private void removeFileByName(String pathToCachedFile) {
    String existingOrMe = getExistingOrMe(pathToCachedFile);
    // we synchronize on the same file name
    synchronized (existingOrMe) {
      File cachedFile = new File(existingOrMe);
      cachedFile.delete();
      filenameMap.remove(existingOrMe);
      boostCache.remove(existingOrMe);
    }
  }

  public void clear() {
    Set<String> names = filenameMap.keySet();
    for (String name : names) {
      removeFileByName(name);
    }
  }


  protected String buildCachedFileName(String region, String key) {
    StringBuffer buf = new StringBuffer(root.getAbsolutePath());
    if (region != null && region.length() > 0)
      buf.append(File.separator).append(region);
    else
      buf.append(File.separator).append(defaultRegion);
    String path = buf.append(File.separatorChar).append(key).append(SUFFIX).toString();
    return path;
  }
}
