package org.jboss.seam.statistics;

import org.jboss.seam.Component;
import org.jboss.seam.ScopeType;
import static org.jboss.seam.annotations.Install.BUILT_IN;
import org.jboss.seam.annotations.*;
import org.jboss.seam.contexts.Contexts;

import java.util.*;

/**
 * All components creating reports can register with this manager and will be called when
 * statistics reports should be generated.
 *
 * @author Sebastian Hennebrueder
 */

@Name("org.jboss.seam.cache.statistics")
@Scope(ScopeType.APPLICATION)
@Install(value = false, precedence = BUILT_IN)
public class StatisticsManager {
	private SortedSet<StatisticsGenerator> statisticsGeneratorSet;

	/**
	 * Creates a report using all registered statistic generators
	 *
	 * @return
	 */
	public String createReport() {
		StringBuilder buf = new StringBuilder();
		for (StatisticsGenerator generator : statisticsGeneratorSet) {
			buf.append(generator.createReport());
		}
		return buf.toString();
	}

	/**
	 * Creates a report only for the passed groups
	 *
	 * @param groups
	 * @return
	 */
	public String createReportForGroup(String... groups) {

		List<String> gs = Arrays.asList(groups);

		StringBuilder buf = new StringBuilder();
		for (StatisticsGenerator generator : statisticsGeneratorSet) {
			if (gs.contains(generator.getStatisticGroupname()))
				buf.append(generator.createReport());
		}
		return buf.toString();
	}

	@Create
	public void start() {
		statisticsGeneratorSet = new TreeSet<StatisticsGenerator>(new StatisticsGeneratorComparator());
	}

	@Destroy
	public void stop() {
		statisticsGeneratorSet.clear();
	}

	/**
	 * Adds a generator to the Set
	 * @param statisticsGenerator
	 */
	public void registerStatisticsGenerator(StatisticsGenerator statisticsGenerator){
		 statisticsGeneratorSet.add(statisticsGenerator);
	}

	/**
	 * Removes a generator. You are reponsible to pass the same instance or create a proper equals/hashcode
	 * @param statisticsGenerator
	 */
	public void unregisterStatisticsGenerator(StatisticsGenerator statisticsGenerator){
		 statisticsGeneratorSet.remove(statisticsGenerator);
	}

	public static StatisticsManager instance() {
		if (!Contexts.isApplicationContextActive()) {
			throw new IllegalStateException("No active application scope");
		}
		return (StatisticsManager) Component.getInstance(StatisticsManager.class, ScopeType.APPLICATION);
	}

	/**
	 * Compares using the group name. It must not be null!
	 */
	public static class StatisticsGeneratorComparator implements Comparator<StatisticsGenerator>{
		public int compare(StatisticsGenerator o1, StatisticsGenerator o2) {
			return o1.getStatisticGroupname().compareTo(o2.getStatisticGroupname());
		}
	}
}
